{ ##
  @FILE                     PJColourPopupMenu.pas
  @COMMENTS                 The component and property editor source code.
  @PROJECT_NAME             Colour Popup Menu
  @PROJECT_DESC             A component that provides a popup menu that displays
                            menu items for the sixteen common colours in a
                            traditional menu style or as a colour palette. A
                            component editor and 16-colour property editor are
                            also included.
  @AUTHOR                   Peter Johnson, LLANARTH, Ceredigion, Wales, UK
  @EMAIL                    peter.johnson@openlink.org
  @WEBSITE                  http://www.delphidabbler.com/
  @COPYRIGHT                 Peter D Johnson, 2001-2003.
  @LEGAL_NOTICE             This component, related component and property
                            editors and source code are placed in the public
                            domain. They may be freely copied and circulated on
                            a not-for-profit basis providing that the code is
                            unmodified and this notice and information about the
                            author and his copyright remains attached to the
                            source code.
  @OWNER                    delphiDabbler
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 06/05/2001
      @COMMENTS             Original version.
    )
    @REVISION(
      @VERSION              1.1
      @DATE                 10/09/2003
      @COMMENTS             Removed design-time code to make compatible with
                            Delphi 6 & 7:
                            + Moved registration on component and property
                              editors to PJColourPopupMenuEd unit and deleted
                              registration routine.
                            + Moved declaration and implementation of property
                              editors to PJColourPopupMenuEd unit.
                            + Removed DsgnIntf unit reference.
    )
  )
}


unit PJColourPopupMenu;

interface


uses
  // Delphi
  SysUtils, Classes, Windows, Graphics, Menus;


type

  {
  TPJColourMenuDisplayStyle:
    Style of menu to display.
  }
  TPJColourMenuDisplayStyle = (
    dsTextOnly,     // displays text only (as standard menu)
    dsColourOnly,   // displays colours only
    dsBoth,         // displays both text and colour
    dsOwnerDraw     // user takes control over display
  );


  {
  TPJColourMenuColumns:
    Number of menu columns to display.
  }
  TPJColourMenuColumns = (
    mc1,            // one column of sixteen colours
    mc2,            // two columns of eight colours each
    mc4,            // four columns of four colours each
    mc8             // eight columns of two colours each
  );


  {
  TPJMeasureColourMenuItemEvent:
    Type of event handler for TPJColourPopupMenu's OnMeasureMenuItem event.
  }
  TPJMeasureColourMenuItemEvent = procedure (Sender: TObject;
    MenuItem: TMenuItem; ACanvas: TCanvas; var Width, Height: Integer;
    Colour: TColor) of object;


  {
  TPJDrawColourMenuItemEvent:
    Type of event handler for TPJColourPopupMenu's OnDrawMenuItem event.
  }
  TPJDrawColourMenuItemEvent = procedure (Sender: TObject; MenuItem: TMenuItem;
    ACanvas: TCanvas; ARect: TRect; Selected: Boolean;
    Colour: TColor) of object;


  {
  TPJCreateColourMenuItemEvent:
    Type of event handler for TPJColourOpupMenu's OnCreateMenuItem event.
  }
  TPJCreateColourMenuItemEvent = procedure (Sender: TObject;
    MenuItem: TMenuItem; Colour: TColor) of object;


  {
  EPJColourPopupMenu:
    Type of exception raised by TPJColourPopupMenu component.
  }
  EPJColourPopupMenu = class(Exception);


  {
  TPJColourPopupMenu:
    A popup menu component that displays either a list of colours or a palette
    of colours.
  }
  TPJColourPopupMenu = class(TPopupMenu)
  private // properties
    fBitmapHeight: Integer;
    fBitmapWidth: Integer;
    fSelectedColour: TColor;
    fMenuColumns: TPJColourMenuColumns;
    fDisplayStyle: TPJColourMenuDisplayStyle;
    fCheckSelectedColour: Boolean;
    fOnSelectColour: TNotifyEvent;
    fOnDrawMenuItem: TPJDrawColourMenuItemEvent;
    fOnMeasureMenuItem: TPJMeasureColourMenuItemEvent;
    fOnCreateMenuItem: TPJCreateColourMenuItemEvent;
  protected // properties
    procedure SetDisplayStyle(const Value: TPJColourMenuDisplayStyle); virtual;
    function GetOwnerDraw: Boolean;
    procedure SetOwnerDraw(const Value: Boolean);
    function GetItems: TMenuItem;
  protected // other
    procedure DoChange(Source: TMenuItem; Rebuild: Boolean); override;
      {Override of method that normally triggers OnChange event - now does
      nothing}
    procedure BuildMenu; virtual;
      {Creates the menu per the component's properties}
    procedure ColourChosen(Sender: TObject); virtual;
      {Event handler for all menu items: updates SelectedColour property and
      triggers OnSelectColour event}
    procedure DrawMenuItem(Sender: TObject; ACanvas: TCanvas; ARect: TRect;
      Selected: Boolean); virtual;
      {Performs custom drawing of colour-only menu items: menu items containing
      text use default drawing}
    procedure MeasureMenuItem(Sender: TObject; ACanvas: TCanvas;
      var Width, Height: Integer); virtual;
      {Returns size of menu item object to be drawn for colour-only menu items:
      menu items containing text use defaults}
    procedure UserDrawMenuItem(Sender: TObject; ACanvas: TCanvas; ARect: TRect;
      Selected: Boolean); virtual;
      {Triggers OnDrawMenuItem event if assigned}
    procedure UserMeasureMenuItem(Sender: TObject; ACanvas: TCanvas;
      var Width, Height: Integer); virtual;
      {Trigger OnMeasurerMenuItem event if assigned}
  public
    constructor Create(AOwner: TComponent); override;
      {Class constructor: sets default property values}
    procedure Popup(X, Y: Integer); override;
      {Creates and displays menu}
  published
    { Re-defined properties }
    property OwnerDraw: Boolean
      read GetOwnerDraw write SetOwnerDraw stored False default False;
      {Changed so that can't be set false when style DisplayStyle is
      dsOwnerDraw. Also changes DisplayStyle to dsOwnerDraw when set true}
    property Items: TMenuItem
      read GetItems;
      {This property is inaccessible in this component: exception is raised if
      any attempt is made to access it}
    { New properties for this component }
    property SelectedColour: TColor
      read fSelectedColour write fSelectedColour default clBlack;
      {The colour last selected on the menu. If no selection has been made, the
      colour is black}
    property DisplayStyle: TPJColourMenuDisplayStyle
      read fDisplayStyle write SetDisplayStyle default dsBoth;
      {The style in which the menu is displayed: text only, colours only, both
      text and colours or owner draw}
    property CheckSelectedColour: Boolean
      read fCheckSelectedColour write fCheckSelectedColour default True;
      {When true the last-selected colour on the menu is highlighted or checked}
    property MenuColumns: TPJColourMenuColumns
      read fMenuColumns write fMenuColumns default mc1;
      {The number of columns that are displayed in menu: 1, 2, 4 or 8}
    property BitmapHeight: Integer
      read fBitmapHeight write fBitmapHeight default 9;
      {The height of the bitmap that displays the colour in the menu. This
      property is ignored unless DisplayStyle property is dsColourOnly}
    property BitmapWidth: Integer
      read fBitmapWidth write fBitmapWidth default 9;
      {The width of the bitmap that displays the colour in the menu. This
      property is ignored unless DisplayStyle property is dsColourOnly}
    property OnSelectColour: TNotifyEvent
      read fOnSelectColour write fOnSelectColour;
      {Event triggered when user selects a colour on the menu. User can get the
      colour that was chosen by examining the SelectedColour property}
    property OnDrawMenuItem: TPJDrawColourMenuItemEvent
      read fOnDrawMenuItem write fOnDrawMenuItem;
      {This event is triggered for each menu item that is to be drawn. User must
      paint the menu item in the event handler. Only triggered when
      DisplayStyle=dsOwnerDraw}
    property OnMeasureMenuItem: TPJMeasureColourMenuItemEvent
      read fOnMeasureMenuItem write fOnMeasureMenuItem;
      {This event is triggered for each menu item that is to be drawn. User must
      return size of menu item in the event handler. Only triggered when
      DisplayStyle=dsOwnerDraw}
    property OnCreateMenuItem: TPJCreateColourMenuItemEvent
      read fOnCreateMenuItem write fOnCreateMenuItem;
      {This event is triggered when colour menu items are being created. The
      menu item can be cutomised before it is displayed}
  end;


implementation


resourcestring
  // Exception message
  sCantAccessItems = 'Can''t access menu items in TPJColourPopupMenu';

const
  // Table mapping colour names to colour identifiers
  cColours: array[0..15] of record
    Name: string;
    Colour: TColor;
  end = (
    (Name: 'Black';   Colour: clBlack;),
    (Name: 'Maroon';  Colour: clMaroon;),
    (Name: 'Green';   Colour: clGreen;),
    (Name: 'Olive';   Colour: clOlive;),
    (Name: 'Navy';    Colour: clNavy;),
    (Name: 'Purple';  Colour: clPurple;),
    (Name: 'Teal';    Colour: clTeal;),
    (Name: 'Grey';    Colour: clGray;),
    (Name: 'Silver';  Colour: clSilver;),
    (Name: 'Red';     Colour: clRed;),
    (Name: 'Lime';    Colour: clLime;),
    (Name: 'Yellow';  Colour: clYellow;),
    (Name: 'Blue';    Colour: clBlue;),
    (Name: 'Fuchsia'; Colour: clFuchsia;),
    (Name: 'Aqua';    Colour: clAqua;),
    (Name: 'White';   Colour: clWhite;)
  );

  // Mapping of TPJColourMenuColumns to actual number of columns
  cMenuCols: array[TPJColourMenuColumns] of Integer = (1, 2, 4, 8);

type

  {
  TColourMenuItem:
    Private custom menu item class for colour menu items: stores colour relating
    to menu item.
  }
  TColourMenuItem = class(TMenuItem)
  private // property
    fColour: TColor;
  public
    property Colour: TColor read fColour write fColour;
      {The colour associated with menu item}
  end;


{ TPJColourPopupMenu }

procedure TPJColourPopupMenu.BuildMenu;
  {Creates the menu per the component's properties}
var
  I: Integer;                 // loops thru all supported colours
  MenuItem: TColourMenuItem;  // new menu item instance being added to menu
  Bmp: TBitmap;               // bitmap instance to be displayed on a menu item
  ColourIdx: Integer;         // index in colours table
  MenuCols: Integer;          // number of columns in menu
  MenuRows: Integer;          // number of rows in menu
  // ---------------------------------------------------------------------------
  function ColourOrder(Idx, Cols, Rows: Integer): Integer;
    {Returns the colour to display at the given column and row to keep colours
    in usual order}
  begin
    Result :=  Cols * (Idx mod Rows) + (Idx div Rows);
  end;
  // ---------------------------------------------------------------------------
begin
  // Do nothing if we're designing
  if csDesigning in ComponentState then
    Exit;
  // Calculate information about number of columns and rows in menu
  MenuCols := cMenuCols[fMenuColumns];
  MenuRows := 16 div MenuCols;
  // Dispose of all previous menu items
  for I := Pred(inherited Items.Count) downto 0 do
    inherited Items[I].Free;
  // Create new menu items
  for I := Low(cColours) to High(cColours) do
  begin
    // Set default property values that can be customised by user
    // calculate index of next colour in colour table
    // (we ensure that colours appear in usual order across columns)
    ColourIdx := ColourOrder(I, MenuCols, MenuRows);
    // create menu item for current colour
    MenuItem := TColourMenuItem.Create(Self);
    if (I mod MenuRows = 0) and (I <> 0) then
      // we're starting a new column
      MenuItem.Break := mbBreak	;
    if fDisplayStyle = dsBoth then
    begin
      // displaying colours with text using default drawing: create bitmap
      Bmp := TBitmap.Create;
      try
        Bmp.Width := 12;   // use hard coded bitmap size: TMenuItem draws this
        Bmp.Height := 12;  // centred in a (hard coded) 16x16 glyph
        Bmp.Canvas.Brush.Color := clMenu;
        Bmp.Canvas.FillRect(Rect(0, 0, Bmp.Width, Bmp.Height));
        Bmp.Canvas.Brush.Color := cColours[ColourIdx].Colour;
        Bmp.Canvas.Pen.Color := clBtnShadow;
        Bmp.Canvas.Rectangle(0, 0, Bmp.Width - 1, Bmp.Height - 1);
        MenuItem.Bitmap := Bmp;
      finally
        Bmp.Free;
      end;
    end;
    if fDisplayStyle in [dsBoth, dsTextOnly, dsOwnerDraw] then
      // displaying colour name using default drawing: store text in caption
      MenuItem.Caption := cColours[ColourIdx].Name;
    if fCheckSelectedColour and
      (cColours[ColourIdx].Colour = fSelectedColour) then
      // we're highlighting check marks and this item is checked, so do it
      MenuItem.Checked := True;
    // record help context
    MenuItem.HelpContext := Self.HelpContext;
    // Record colour in menu item's tag
    MenuItem.Colour := cColours[ColourIdx].Colour;
    // Allow user to cutomise menu
    if Assigned(fOnCreateMenuItem) then
      fOnCreateMenuItem(Self, MenuItem, MenuItem.Colour);
    // Restore event handlers as required
    // set measure and drawing items if owner draw
    case DisplayStyle of
      dsColourOnly:
      begin
        // component draws colour-only style
        MenuItem.OnDrawItem := DrawMenuItem;
        MenuItem.OnMeasureItem := MeasureMenuItem;
      end;
      dsOwnerDraw:
      begin
        // user draws using events
        MenuItem.OnDrawItem := UserDrawMenuItem;
        MenuItem.OnMeasureItem := UserMeasureMenuItem;
      end;
      else
        // no owner draw events
        MenuItem.OnDrawItem := nil;
        MenuItem.OnMeasureItem := nil;
    end;
    // intercept menu item's OnClick event
    MenuItem.OnClick := ColourChosen;
    // Add new menu item to menu
    inherited Items.Add(MenuItem);
  end;
end;

procedure TPJColourPopupMenu.ColourChosen(Sender: TObject);
  {Event handler for all menu items: updates SelectedColour property and
  triggers OnSelectColour event}
begin
  // Record selected colour and trigger OnSelectColour event
  fSelectedColour := (Sender as TColourMenuItem).Colour;
  if Assigned(fOnSelectColour) then fOnSelectColour(Self);
end;

constructor TPJColourPopupMenu.Create(AOwner: TComponent);
  {Class constructor: sets default property values}
begin
  inherited Create(AOwner);
  fDisplayStyle := dsBoth;
  fCheckSelectedColour := True;
  fMenuColumns := mc1;
  fBitmapHeight := 9;
  fBitmapWidth := 9;
  fSelectedColour := clBlack;
end;

procedure TPJColourPopupMenu.DoChange(Source: TMenuItem; Rebuild: Boolean);
  {Override of method that normally triggers OnChange event - now does nothing}
begin
  {Do nothing};
end;

procedure TPJColourPopupMenu.DrawMenuItem(Sender: TObject;
  ACanvas: TCanvas; ARect: TRect; Selected: Boolean);
  {Performs custom drawing of colour-only menu items: menu items containing text
  use default drawing}
  // ---------------------------------------------------------------------------
  function CreateBackgroundBmp: TBitmap;
    {Returns a new bitmap object that is used to draw background to checked
    bitmap items}
  var
    X, Y: Integer;      // loop thru X and Y pixels of bitmap
  begin
    // Create bitmap of required brush size
    Result := TBitmap.Create;
    Result.Width := 8;
    Result.Height := 8;
    // Fill background with required colour
    Result.Canvas.Brush.Color := clMenu;
    Result.Canvas.FillRect(Rect(0, 0, 8, 8));
    // Create chequerboard pattern using button highlight colour
    for X := 0 to 7 do
      for Y := 0 to 7 do
        if (Y mod 2) = (X mod 2) then
          Result.Canvas.Pixels[X, Y] := clBtnHighlight;
  end;
  // ---------------------------------------------------------------------------
var
  MenuItem: TColourMenuItem;  // reference to menu item being drawn
  BrushBmp: TBitmap;          // brush bitmap used to highlight checked items
begin
  // Record reference to menu item being drawn, it's text and related colour
  MenuItem := Sender as TColourMenuItem;
  // Adjust size of area to be drawn
  InflateRect(ARect, -1, 0);
  // Draw background to bitmap
  BrushBmp := nil;        // this object is used to draw background if "checked"
  try
    if fCheckSelectedColour and MenuItem.Checked and not Selected then
    begin
      // menu item is checked and we're flagging as such: use background brush
      BrushBmp := CreateBackgroundBmp;
      ACanvas.Brush.Bitmap := BrushBmp;
    end
    else
    begin
      // menu item not to have checked highlight, set appropriate brush colour
      ACanvas.Brush.Bitmap := nil;
      ACanvas.Brush.Color := clMenu;
    end;
    // actually draw the background
    ACanvas.FillRect(ARect);
  finally
    BrushBmp.Free;
  end;
  // Draw the raised edge around the bitmap area as raised or lowered
  if fCheckSelectedColour and MenuItem.Checked then
    DrawEdge(ACanvas.Handle, ARect, BDR_SUNKENOUTER, BF_RECT)
  else if Selected then
    DrawEdge(ACanvas.Handle, ARect, BDR_RAISEDINNER, BF_RECT);
  // Draw actual colour
  // adjust rectangle size for drawing colour
  InflateRect(ARect, -4, -4);
  // now draw it
  ACanvas.Pen.Color := clBtnShadow;
  ACanvas.Brush.Color := MenuItem.Colour;
  ACanvas.Rectangle(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom);
end;

function TPJColourPopupMenu.GetItems: TMenuItem;
  {Replacement read access method for Items property: raises exception at run
  time}
begin
  if csDesigning in ComponentState then
    Result := inherited Items
  else
    raise EPJColourPopupMenu.Create(sCantAccessItems);
end;

function TPJColourPopupMenu.GetOwnerDraw: Boolean;
  {Replacement read access method for OwnerDraw property: uses DisplayStyle
  value}
begin
  Result := fDisplayStyle = dsOwnerDraw;
end;

procedure TPJColourPopupMenu.MeasureMenuItem(Sender: TObject;
  ACanvas: TCanvas; var Width, Height: Integer);
  {Returns size of menu item object to be drawn for colour-only menu items: menu
  items containing text use defaults}
begin
  Height := fBitmapHeight + 10;     // adjust so that smallest height shows
  Width := fBitmapWidth;
end;

procedure TPJColourPopupMenu.Popup(X, Y: Integer);
  {Creates and displays menu}
begin
  BuildMenu;
  inherited Popup(X, Y);
end;

procedure TPJColourPopupMenu.SetDisplayStyle(
  const Value: TPJColourMenuDisplayStyle);
  {Write access method for DisplayStyle property: stores value and also sets
  inherited OwnerDraw property to required value depending on DisplayStyle}
begin
  if fDisplayStyle = Value then Exit;
  fDisplayStyle := Value;
  inherited OwnerDraw := Value in [dsColourOnly, dsOwnerDraw];
end;

procedure TPJColourPopupMenu.SetOwnerDraw(const Value: Boolean);
  {Replacement write access method for OwnerDraw property: sets DisplayStyle to
  dsOwnerDraw if the OwnerDraw property is set true}
begin
  if Value = True then
    SetDisplayStyle(dsOwnerDraw);
end;

procedure TPJColourPopupMenu.UserDrawMenuItem(Sender: TObject;
  ACanvas: TCanvas; ARect: TRect; Selected: Boolean);
  {Triggers OnDrawMenuItem event if assigned}
begin
  if Assigned(fOnDrawMenuItem) then
    fOnDrawMenuItem(Self, Sender as TMenuItem, ACanvas, ARect, Selected,
      (Sender as TColourMenuItem).Colour);
end;

procedure TPJColourPopupMenu.UserMeasureMenuItem(Sender: TObject;
  ACanvas: TCanvas; var Width, Height: Integer);
  {Trigger OnMeasurerMenuItem event if assigned}
begin
  if Assigned(fOnMeasureMenuItem) then
    fOnMeasureMenuItem(Self, Sender as TMenuItem, ACanvas, Width, Height,
      (Sender as TColourMenuItem).Colour);
end;

end.
